/*
 * advanced-vanadiel-clock.js
 *
 * Copyright (c) 2004 by Daniel Potter.  All rights reserved.
 *
 * This file is part of the Vana'diel Clock.
 *
 * The Vana'diel Clock is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The Vana'diel Clock is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Vana'diel Clock; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * A more complicated clock, this handles the clock for the "advanced" display
 * and contains two ticks - Earth time and Vanadiel time.
 */

// Name of Vana'diel days - note that these days are the value we give to the
// "day" attribute on our time marker to set the day icon.

const vanaDays = [ "fire", "earth", "water", "wind", "ice", "thunder", "light", "dark" ];

var vanaLocalNames = new Array();

// Name of moon phases, used to look up the phase name and set the moon icon
const moonPhases = [ "new", "waxingcrescent", "firstquarter", "waxinggibbous", "full", "waninggibbous", "lastquarter", "waningcrescent" ];

var moonPhaseLocalNames = new Array();
var moonPhaseLocalWaxing;
var moonPhaseLocalWaning;

// Offset (in Earth milliseconds) to convert Earth time to Vanadiel 1/25ths of
// a millisecond.
var vanadielOffset = 0;

/*** AIRSHIP CONSTANTS ***/
// These are based on the Mithra Pride airship timer found at
// http://www.mithrapride.org/vana_time/airshiptable.html
//
// I haven't checked these yet myself - the Kazham and Windurst times
// seem to be a little off.

// Amount of time the airship takes, total, between trips (6 Vana'diel hours,
// approximately 15 Earth minutes - note this includes boarding time)
var airshipDelay = 6 * 3600000;

// Offsets from "base time" to the various airship departure times.

var airshipTimeDelays = [
    4 * 3600000 + 10 * 60000,  1 * 3600000 + 10 * 60000,  // San d'Oria in/out
    1 * 3600000 + 10 * 60000, 4 * 3600000 + 10 * 60000,   // Bastok in/out
    5 * 3600000 + 45 * 60000, 2 * 3600000 + 40 * 60000,   // Windurst in/out
    2 * 3600000 + 40 * 60000, 5 * 3600000 + 35 * 60000 ]; // Kazham in/out

// Cell names to place calculate airship times into
var airshipIds = [
    "sandoria.inbound", "sandoria.outbound",
    "bastok.inbound", "bastok.outbound",
    "windurst.inbound", "windurst.outbound",
    "kazham.inbound", "kazham.outbound"
];
var airshipElements = [
    null, null, null, null, null, null, null, null
];

// Object used for representing the races of Vana'diel (used for the RSE
// timer)
function VanaRace(name, icon) {
    this.name = name;
    this.icon = icon;
}

// The various RSE races, in RSE order

var RSERaces = [
    new VanaRace('elvaan.male', 'elvaan-m-icon'),
    new VanaRace('elvaan.female', 'elvaan-f-icon'),
    new VanaRace('tarutaru.male', 'tarutaru-m-icon'),
    new VanaRace('tarutaru.female', 'tarutaru-f-icon'),
    new VanaRace('mithra', 'mithra-icon'),
    new VanaRace('galka', 'galka-icon'),
    new VanaRace('hume.male', 'hume-m-icon'),
    new VanaRace('hume.female', 'hume-f-icon')
];

var RSEActive = "?";
var RSETodayAt = "";
var RSETomorrowAt = "";

var RSELocations = [
    "rse.ordelle",
    "rse.gusgen",
    "rse.shakhrami",
];

var earthMonths = [
    "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"
];

var earthMonthsAbbrv = new Array(earthMonths.length); // in case we get a new month, or - shut up.

var earthDaysInMonth = [
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
];

var earthAM = "A";
var earthPM = "P";

var earth24HourClock = false;

{
    // Load our localized strings
    var strBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
    var strings = strBundleService.createBundle("chrome://vanadielclock/locale/clock.properties");

    for (var i in vanaDays) {
	vanaLocalNames[i] = strings.GetStringFromName("day." + vanaDays[i]);
    }
    for (var i in moonPhases) {
	moonPhaseLocalNames[i] = strings.GetStringFromName("moon." + moonPhases[i]);
    }
    for (var i in RSERaces) {
	try {
	    RSERaces[i].name = strings.GetStringFromName("race." + RSERaces[i].name);
	} catch (ex) {
	    alert(RSERaces[i].name + " not localized!\n\nThis message not localized!");
	}
    }
    for (var i in RSELocations) {
	RSELocations[i] = strings.GetStringFromName(RSELocations[i]);
    }
    for (var i in earthMonths) {
	earthMonthsAbbrv[i] = strings.GetStringFromName("month.abbrv." + earthMonths[i]);
	earthMonths[i] = strings.GetStringFromName("month." + earthMonths[i]);
    }
    moonPhaseLocalWaxing = strings.GetStringFromName("moon.waxing");
    moonPhaseLocalWaning = strings.GetStringFromName("moon.waning");
    RSEActive = strings.GetStringFromName("rse.active");
    RSETodayAt = strings.GetStringFromName("rse.today");
    RSETomorrowAt = strings.GetStringFromName("rse.tomorrow");
    earthAM = strings.GetStringFromName("clock.am");
    earthPM = strings.GetStringFromName("clock.pm");

    // Calculate the offset to convert to Vana'diel time for the current time zone

    // I, unfortunately, have no idea who originally came up with the following.
    // Note that it has since been converted to GMT.
    
    // Earth time:       2002/06/23 15:00:00 GMT
    // Vana'diel time:   0898/02/01 00:00:00 (VST?)

    var vanaTime  = (898 * 360 + 30) * 24 * 60 * 60 * 1000 / 25;
    var baseDay = new Date();
    baseDay.setUTCFullYear(2002, 5, 23); // Set date to 2002-06-23
    baseDay.setUTCHours(15, 0, 0, 0);    // Set time to 15:00:00.0000

    vanadielOffset = baseDay.getTime() - vanaTime;

    var prefs = Components.classes["@mozilla.org/preferences-service;1"].
               getService(Components.interfaces.nsIPrefBranch);

    try {
	earth24HourClock = prefs.getBoolPref("vanadiel.clock.earth.24hour");
    } catch (ex) {
	// use default
    }

    // add an onload listener
    addEventListener("load",
	    function(event) { createRSERows();startClock() },
	    false);
}

var vanadielMoonIcon = null;
var vanadielMoonLabel = null;
var vanadielDayIcon = null;
var vanadielDateLabel = null;
var vanadielTimeLabel = null;
var earthTimeLabel = null;
var vanadielClockInterval = null;
var earthClockInterval = null;
var vanadielTime = -1;
var vanadielLastDay = -1;
var vanadielLastMinute = -1;
var earthLastSecond = -1;

var airshipTimes = [ 0, 0,
		     0, 0,
		     0, 0,
		     0, 0 ];

// Starts up the advanced clock.
function startClock() {
    if (vanadielClockInterval == null) {
	vanadielDayIcon = document.getElementById("day-icon");
	vanadielDateLabel = document.getElementById("date-label");
	vanadielTimeLabel = document.getElementById("timer-label");
	vanadielMoonIcon = document.getElementById("moon-phase-icon");
	vanadielMoonLabel = document.getElementById("moon-phase-label");
	earthTimeLabel = document.getElementById("earth-label");
	for (var i in airshipElements) {
	    airshipElements[i] = document.getElementById(airshipIds[i]);
	}
	// immediately update the clock
	vanadielLastDay = -1;
	updateVanadielClock();
	updateEarthClock();
	vanadielClockInterval = setInterval("updateVanadielClock()");
	earthClockInterval = setInterval("updateEarthClock()");
    }
}

// Stopping the clock is fairly pointless, but - do it anyway.
function stopClock() {
    if (vanadielClockInterval != null) {
	clearInterval(vanadielClockInterval);
	vanadielClockInterval = null;
    }
    if (earthClockInterval != null) {
	clearInterval(earthClockInterval);
	earthClockInterval = null;
    }
}

function updateVanadielClock() {
    // Vana'diel time runs 25 times faster than our time - this is the time
    // in "Vana'diel milliseconds" actually.
    vanadielTime = (new Date().getTime() - vanadielOffset) * 25;
    var vanadielMinute = Math.floor(vanadielTime / 60000) % 60;
    // If the minute hasn't changed yet, abort - nothing's changed
    if (vanadielMinute == vanadielLastMinute)
	return;
    vanadielLastMinute = vanadielMinute;
    // We only display the day, hour, and minute in the little clock
    var vanadielDay = Math.floor(vanadielTime / 86400000);
    // Only update day related stuff on - well, a new day
    if (vanadielDay != vanadielLastDay) {
	vanadielLastDay = vanadielDay;
	var vanadielYear = Math.floor(vanadielDay / 360);
	var vanadielMonth = Math.floor((vanadielDay % 360) / 30) + 1;
	var vanadielDayMonth = (vanadielDay % 30) + 1;
	var vanadielDayWeek = vanadielDay % 8;
	// calculate the current moon phase - unlike the current
	// day of the week, this isn't nice and happily rooted at 0
	// moonphase = 0 - full moon
	// moonphase = 42 - new moon
	var moonphase = (vanadielDay + 26) % 84;
	// translate that value into a percentage value
	var moonpercent = ((42 - moonphase) * 100) / 42;
	// negative values are "waxing" and backwards
	var waxing = false;
	if (moonpercent < 0) {
	    waxing = true;
	    moonpercent = Math.abs(moonpercent);
	}
	// Now make it an interger
	moonpercent = Math.floor(moonpercent + 0.5);
	// Translate the phase into one of the eight phases.
	//
	// To see how the "phase counter" stacks up to a percent, run the
	// following Perl 1-liner:
	//
	// perl -e 'for($p=0;$p<84;$p++){print "$p = ".int((abs((42-$p)*100/42)+0.5))."\% (wa".($p<42?"n":"x")."ing)\n";}'
	//
	// Simple, huh?
	//
	// === MOON PHASES ===
	// 80- 2 == 90%(Wax)  - 95%(Wane) - Full
	//  3-16 == 93%(Wane) - 62%(Wane) - Waning Gibbous
	// 17-23 == 60%(Wane) - 45%(Wane) - Last Quarter
	// 24-37 == 43%(Wane) - 12%(Wane) - Waning Crescent
	// 38-44 == 10%(Wane) -  5%(Wax)  - New Moon
	// 45-58 ==  7%(Wax)  - 38%(Wax)  - Waxing Crescent
	// 59-65 == 40%(Wax)  - 55%(Wax)  - First Quarter
	// 66-79 == 57%(Wax)  - 88%(Wax)  - Waxing Gibbous
	var basicMoonPhase = 4; // full
	if (moonphase < 3) {
	    basicMoonPhase = 4; // full
	} else if (moonphase < 17) {
	    basicMoonPhase = 5; // waning gibbous
	} else if (moonphase < 24) {
	    basicMoonPhase = 6; // last quarter
	} else if (moonphase < 38) {
	    basicMoonPhase = 7; // waning crescent
	} else if (moonphase < 45) {
	    basicMoonPhase = 0; // new
	} else if (moonphase < 59) {
	    basicMoonPhase = 1; // waxing crescent
	} else if (moonphase < 66) {
	    basicMoonPhase = 2; // first quarter
	} else if (moonphase < 80) {
	    basicMoonPhase = 3; // waxing gibbous
	} else {
	    basicMoonPhase = 4; // full
	}
	var date = vanaLocalNames[vanadielDayWeek] + ", " + vanadielDayMonth + " / ";
	date += vanadielMonth + " / ";
	date += vanadielYear;
	//dump("Calculated moon at " + moonpercent + "% " + (waxing ? "waxing" : "waning"));
	//dump("\nPhase counter is " + moonphase + ", basic phase is " + basicMoonPhase + " (" + moonPhases[basicMoonPhase] + ")\n");
	vanadielMoonIcon.setAttribute("phase", moonPhases[basicMoonPhase]);
	var moonPhaseLabel = moonpercent + "% " + moonPhaseLocalNames[basicMoonPhase];
	if (basicMoonPhase == 0 || basicMoonPhase == 4) {
	    // add "waxing" or "waning" to this if you can't tell
	    // (ie, when the moon is new or full, so you know if it's
	    // moving to/from the "true" full/new state)
	    if (moonphase != 0 && moonphase != 42) {
		// But don't bother showing waxing/waning when the moon
		// really is completely full or completely new - it should
		// (hopefully) be obvious that it's going to move in the
		// other direction.
		moonPhaseLabel += " (" + (waxing ? moonPhaseLocalWaxing : moonPhaseLocalWaning) + ")";
	    }
	}
	vanadielMoonIcon.setAttribute("tooltiptext", moonPhaseLabel);
	vanadielMoonLabel.setAttribute("value", moonPhaseLabel);
	vanadielDayIcon.setAttribute("tooltiptext", date);
	vanadielDayIcon.setAttribute("day", vanaDays[vanadielDayWeek]);
	vanadielDateLabel.setAttribute("value", date);
	vanadielTimeLabel.setAttribute("tooltiptext", date);
	updateRSE();
    }
    var vanadielHour = Math.floor(vanadielTime / 3600000) % 24;
    var label = vanadielHour + ":";
    if (vanadielMinute < 10) {
	label += "0";
    }
    label += vanadielMinute;
    vanadielTimeLabel.setAttribute("value", label);
    // update the airship times (I guess)
    for (var i in airshipTimes) {
	airshipTimes[i] = (Math.floor(vanadielTime / airshipDelay) * airshipDelay) + airshipTimeDelays[i];
	if (airshipTimes[i] < vanadielTime)
	    airshipTimes[i] += airshipDelay;
    }
}

function updateEarthClock() {
    var earthTime = new Date();
    var sec = earthTime.getSeconds();
    if (earthLastSecond == sec)
	return;
    earthLastSecond == sec;
    var min = earthTime.getMinutes();

    var label = formatEarthDate(earthTime) + " " +
		formatEarthTime(earthTime);
    earthTimeLabel.setAttribute("value", label);
    // Update airship delays.
    for (var i in airshipTimes) {
	sec = Math.floor((airshipTimes[i] - vanadielTime) / 25000);
	min = Math.floor(sec/60);
	sec = sec - min * 60;
	label = min + ":";
	if (sec < 10) {
	    label += "0";
	}
	label += sec;
	airshipElements[i].setAttribute("value", label)
    }
}

function createRSERows() {
    var rows = document.getElementById("rse-rows");
    for (var i = 0; i < 8; i++) {
	var row = document.createElement("row");
	rows.appendChild(row);
	row.setAttribute("align", "center");
	var image = document.createElement("image");
	image.setAttribute("id", "rse-icon-" + i);
	row.appendChild(image);
	var label = document.createElement("label");
	label.setAttribute("id", "rse-race-label-" + i);
	row.appendChild(label);
	var label = document.createElement("label");
	label.setAttribute("id", "rse-location-label-" + i);
	row.appendChild(label);
	var label = document.createElement("label");
	label.setAttribute("id", "rse-time-label-" + i);
	row.appendChild(label);
    }
}

// Update the RSE timers.
// Only called on new days.
function updateRSE() {
    // Create a list of RSE rows
    // Figure out which week this is...
    var vanadielWeek = Math.floor(vanadielTime / 691200000);
    // Now start calculating the next 8 weeks (so we go through all races)
    var today = new Date();
    for (var i = 0; i < 8; i++,vanadielWeek++) {
	var race = vanadielWeek % 8;
	var location = vanadielWeek % 3;
	var startTime = vanadielWeek * 691200000;
	var startDate = new Date(Math.floor(startTime/25)+vanadielOffset);
	var label;
	if (i == 0) {
	    label = RSEActive;
	} else {
	    var day = today.getDate();
	    var rseday = startDate.getDate();
	    if (startDate.getMonth() > today.getMonth()) {
		// can't compare them directly, add the number of days in
		// today's month to the rse month
		if (today.getMonth() >= 1) {
		    if (today.getFullYear() % 4 == 0 &&
			(today.getFullYear() % 100 != 0 || today.getFullYear() % 1000 == 0)) {
			rseday++;
		    }
		}
		rseday += earthDaysInMonth[today.getMonth()];
		// ... ah, no leap years, uniform months... I like Vana'diel's
		// calendar (*sigh*)
	    }
	    if (day == rseday) {
		// still today, just display the time
		label = RSETodayAt + " " + formatEarthTime(startDate, false);
	    } else if (day + 1 == rseday) {
		label = RSETomorrowAt + " " + formatEarthTime(startDate, false);
	    } else {
		label = formatEarthDate(startDate) + " " + formatEarthTime(startDate, false);
	    }
	}
	document.getElementById("rse-icon-" + i).setAttribute("class", RSERaces[race].icon);
	document.getElementById("rse-race-label-" + i).setAttribute("value", RSERaces[race].name);
	document.getElementById("rse-location-label-" + i).setAttribute("value", RSELocations[location]);
	document.getElementById("rse-time-label-" + i).setAttribute("value", label);
    }
}

// Returns a string formatted in Earth time (based on the user's settings)
function formatEarthDate(time) {
    return time.getFullYear() + ", " +
	    earthMonths[time.getMonth()] + " " +
	    time.getDate();
}

// Returns a string formatted in Earth time (based on the user's settings)
function formatEarthTime(time, showSeconds) {
    var result = time.getHours();
    var am = false;
    if (!earth24HourClock) {
	if (result > 12) {
	    result -= 12;
	} else {
	    am = true;
	}
    }
    var t = time.getMinutes();
    result += (t < 10 ? ":0" : ":") + t;
    if (showSeconds == null || showSeconds) {
	t = time.getSeconds();
	result += (t < 10 ? ":0" : ":") + t;
    }
    if (!earth24HourClock) {
	result += " " + (am ? earthAM : earthPM);
    }
    return result;
}
